/*
	Tytu: Boost::pool
	Autor: Igor Rutkowski
	Grupa: G1EIK
*/

/*
	Standardowy operator new jest rozwizaniem uniwersalnym i jako takie
obciony jest wieloma kompromisami. Nie jest to rozwizanie najwydajniejsze,
zwaszcza dla maych obiektw. Pewnym antidotum jest wstpne zaalokowanie
wikszego obszaru pamici i dzielenie jej na mniejsze fragmenty. Zadanie to moe
uatwi nam biblioteka boost::pool. Jest ona stosunkowo prosta, nie udostpnia 
uytkownikowi zbyt wielu klas. Jej uytkowanie sprowadza si do doczenia 
plikw nagwkowych.

*/

#include <boost/pool/pool.hpp>
#include <boost/pool/object_pool.hpp>
#include <boost/pool/singleton_pool.hpp>
#include <ctime>
#include <iostream>

void speed_test(void);
using namespace std;

//Prosta struktura, pozwalajca ledzi wywoywanie konstruktorw i destruktorw.
struct X{
	X()  {	cout << "Konstruktor    "<<endl;}
	~X() {	cout << "Destruktor    "<<endl;}
};
//Dla boost::singleton_pool
struct Tag { };


int main(int argc, char * argv[]){
	int* int_pointer; 
	int* int_tab;
	X* X_pointer1;
	X* X_pointer2;

	//Prosty alokator pamici. Obsuga bdw przez zwrcenie wartoci NULL.
	//Nie jest przystosowany do obsugi obiektw.
	{
		boost::pool<> p(sizeof(int));
		//Jeden element
		int_pointer = reinterpret_cast<int *>(p.malloc());
		//Tablica elementow
		int_tab = reinterpret_cast<int *>(p.ordered_malloc(5));
		//Zwolnienie czesci tablicy
		p.free(int_tab+2,3);
		//Zwolnienie jednego elementu
		p.free(int_pointer);
		//Zwolnienie wszystkich blokw
		//p.purge_memory();
	} //	Nie jest to jednak konieczne, gdy nastpuje automatycznie wraz z
	  //wywoaniem destruktora obiektu pool.
	cout << "Boost::pool dla struktury"<<endl;
	{
		boost::pool<> p(sizeof(X));
		X_pointer1 = reinterpret_cast<X *>(p.malloc());
		X_pointer2 = reinterpret_cast<X *>(p.malloc());
		p.free(X_pointer1);
	}
	//W adnym momencie nie nastpio wywoanie konstruktora ani destruktora

	cout << "Boost::object_pool"<<endl;
	//Alokator przystosowany do zarzdzania obiektami ale dziaajcy rwnie z typami prostymi.
	//Obsuga bdw j.w. Nie przewidziano moliwoci alokowania tablic.
	{
		boost::object_pool<X> p;
		cout<<"malloc():"<<endl;
		X_pointer1 = p.malloc();
		cout<<"construct():"<<endl;
		X_pointer2 = p.construct();
		cout<<"free():"<<endl;
		p.free(X_pointer1);
		cout<<"destroy():"<<endl;
		p.destroy(X_pointer2);
		X_pointer1 = p.malloc();
		X_pointer2 = p.construct();
		cout <<"Wywolanie destruktora object_pool"<<endl;
	}
	//Destruktor wywoano dla obiektw stworzonych zarwno przez malloc() jak i construct()
	//Moe to by przyczyn istotnych bdw
	cout<<endl;

	//Istnieje rwnie wersja boost::pool przystosowana do wielowtkowoci.
	//Implementuje ona wzorzec programowy Singleton. Obsuga bdw j.w.
	//Struktura Tag jedynie rozrnia nam obiekty, gdy moliwe jest istnienie kilku
	//alokatorw dla tego samego typu.
	typedef boost::singleton_pool<Tag, sizeof(int)>  single_pool;
	{
		int * intp = reinterpret_cast<int *>(single_pool::malloc());
		single_pool::free(intp);
		int_pointer = reinterpret_cast<int *>(single_pool::malloc());
		
	} //Teraz moemy zwolni caa pami? przydzielon - nie odbywa sie to automatycznie wraz
	// z wyjciem z bloku!
	single_pool::purge_memory();
	//Poza tym, uywamy tego alokatora tak samo jak boost::pool - udostpnia on identyczn funkcjonalno.


	speed_test();
	return 0;
}

//Srednia z 10 uruchomie? (kompilacja : g++ -O3)
/*
Malloc i free : 779
New i delete : 1034
Boost::pool dla int
Malloc i free : 30
Boost::object_pool dla int
Construct i destroy : 79
Malloc i free : 71
*/
//Przy braku optymalizacji (-O0) wyniki dla bibliotek boost s znacznie gorsze.
//Rnice w wydajnoci s bardzo wyrane, dlatego warto rozway uywanie tej biblioteki.



void speed_test(void){
	//Dla systemu MS Windows
	const int div =1;
	//Dla Linuxa
	//const int div=1000;
	clock_t time1,time2;
	size_t i,size = 1024*1024*10;
	int* p;
	boost::object_pool<int> obj_pool;
	boost::pool<> pool(sizeof(int));

	cout<<"Test szybkosci"<<endl;
	time1=clock();
	for (i = 0; i < size; ++i)
	{
		p = reinterpret_cast<int *>(malloc(sizeof(int)));
		free(p);
	}
	time2=clock();
	std::cout << "Malloc i free : " << (time2 - time1)/div << std::endl;

	time1=clock();
	for (i = 0; i < size; ++i)
	{
		p = new int;
		delete p;
	}
	time2=clock();
	std::cout << "New i delete : " << (time2 - time1)/div << std::endl;

	std::cout << "Boost::pool dla int\n";

	time1=clock();
	for (i = 0; i < size; ++i)
	{
		p = reinterpret_cast<int *>(pool.malloc());
		pool.free(p);
	}
	time2=clock();
	std::cout <<  "Malloc i free : "<< (time2 - time1)/div << std::endl;	

	std::cout << "Boost::object_pool dla int\n";

	
	time1=clock();
	for (i = 0; i < size; ++i)
	{
		p = obj_pool.construct();
		obj_pool.destroy(p);
	}
	time2=clock();
	std::cout << "Construct i destroy : " << (time2 - time1)/div << std::endl;

	time1=clock();
	for (i = 0; i < size; ++i)
	{
		p = obj_pool.malloc();
		obj_pool.free(p);
	}
	time2=clock();
	std::cout <<  "Malloc i free : "<< (time2 - time1)/div << std::endl;
}
